"""Model path tracing tests."""

import pytest
from networkx.exception import NetworkXNoPath
from finesse.components import Beamsplitter, Laser, Mirror, Cavity
from finesse.detectors import PowerDetector


@pytest.fixture
def fabry_perot(model):
    """Fixture for constructing a Fabry-Perot cavity."""
    model.add([Laser("L0"), Mirror("ITM", Rc=-0.6), Mirror("ETM", Rc=0.6)])
    model.connect(model.L0.p1, model.ITM.p1)
    model.connect(model.ITM.p2, model.ETM.p1, L=1)
    model.add(Cavity("cav", model.ITM.p2.o, model.ITM.p2.i))
    model.add(PowerDetector("REFL", model.ITM.p1.o))

    return model


@pytest.fixture
def simple_michelson(model):
    """Fixture for constructing a simple Michelson interferometer."""
    model.add([Laser("L0"), Beamsplitter("BS"), Mirror("MX"), Mirror("MY")])
    model.connect(model.L0.p1, model.BS.p1)
    model.connect(model.BS.p2, model.MY.p1, L=1)
    model.connect(model.BS.p3, model.MX.p1, L=1)
    model.add(PowerDetector("AS", model.BS.p4.o))

    return model


def test_fullpath_laser_to_mirror(model):
    """Test path finding from a single laser to a single mirror."""
    model.add([Laser("L0"), Mirror("M")])
    model.connect(model.L0.p1, model.M.p1)

    # Transmitted through mirror.
    result_transmit = model.path(model.L0.p1.o, model.M.p2.o).data
    target_transmit = [
        (model.L0.p1.o, model.L0.p1.o.space),
        (model.M.p1.i, model.M),
        (model.M.p2.o, None),
    ]
    assert result_transmit == target_transmit

    # Reflected from mirror.
    result_reflect = model.path(model.L0.p1.o, model.M.p1.o).data
    target_reflect = [
        (model.L0.p1.o, model.L0.p1.o.space),
        (model.M.p1.i, model.M),
        (model.M.p1.o, model.L0.p1.o.space),
    ]
    assert result_reflect == target_reflect

    # Impossible path.
    with pytest.raises(NetworkXNoPath):
        model.path(model.L0.p1.o, model.M.p2.i)


def test_fullpath_fabry_perot(fabry_perot):
    """Test path finding from a single laser to a Fabry-Perot cavity."""
    # Transmitted through Fabry-Perot.
    result_transmit_all = fabry_perot.path(
        fabry_perot.L0.p1.o, fabry_perot.ETM.p2.o
    ).data
    target_transmit_all = [
        (fabry_perot.L0.p1.o, fabry_perot.L0.p1.o.space),
        (fabry_perot.ITM.p1.i, fabry_perot.ITM),
        (fabry_perot.ITM.p2.o, fabry_perot.ITM.p2.o.space),
        (fabry_perot.ETM.p1.i, fabry_perot.ETM),
        (fabry_perot.ETM.p2.o, None),
    ]
    assert result_transmit_all == target_transmit_all

    # Reflected back to ITM from ETM.
    result_half_roundtrip = fabry_perot.path(
        fabry_perot.L0.p1.o, fabry_perot.ITM.p2.i
    ).data
    target_half_roundtrip = [
        (fabry_perot.L0.p1.o, fabry_perot.L0.p1.o.space),
        (fabry_perot.ITM.p1.i, fabry_perot.ITM),
        (fabry_perot.ITM.p2.o, fabry_perot.ITM.p2.o.space),
        (fabry_perot.ETM.p1.i, fabry_perot.ETM),
        (fabry_perot.ETM.p1.o, fabry_perot.ETM.p1.o.space),
        (fabry_perot.ITM.p2.i, fabry_perot.ITM),
    ]
    assert result_half_roundtrip == target_half_roundtrip


def test_fullpath_fabry_perot_with_via_nodes(fabry_perot):
    """Test path finding from a single laser to a Fabry-Perot cavity via specified
    nodes."""
    # Full roundtrip.
    result_roundtrip = fabry_perot.path(
        fabry_perot.L0.p1.o, fabry_perot.ETM.p2.o, via_node=fabry_perot.ITM.p2.i
    ).data
    target_roundtrip = [
        (fabry_perot.L0.p1.o, fabry_perot.L0.p1.o.space),
        (fabry_perot.ITM.p1.i, fabry_perot.ITM),
        (fabry_perot.ITM.p2.o, fabry_perot.ITM.p2.o.space),
        (fabry_perot.ETM.p1.i, fabry_perot.ETM),
        (fabry_perot.ETM.p1.o, fabry_perot.ETM.p1.o.space),
        (fabry_perot.ITM.p2.i, fabry_perot.ITM),
        (fabry_perot.ITM.p2.o, fabry_perot.ITM.p2.o.space),
        (fabry_perot.ETM.p1.i, fabry_perot.ETM),
        (fabry_perot.ETM.p2.o, None),
    ]
    assert result_roundtrip == target_roundtrip

    # Reflection from ITM but via ETM.
    result_ITM_refl = fabry_perot.path(
        fabry_perot.L0.p1.o, fabry_perot.ITM.p1.o, via_node=fabry_perot.ETM.p1.i
    ).data
    target_ITM_refl = [
        (fabry_perot.L0.p1.o, fabry_perot.L0.p1.o.space),
        (fabry_perot.ITM.p1.i, fabry_perot.ITM),
        (fabry_perot.ITM.p2.o, fabry_perot.ITM.p2.o.space),
        (fabry_perot.ETM.p1.i, fabry_perot.ETM),
        (fabry_perot.ETM.p1.o, fabry_perot.ETM.p1.o.space),
        (fabry_perot.ITM.p2.i, fabry_perot.ITM),
        (fabry_perot.ITM.p1.o, fabry_perot.ITM.p1.o.space),
    ]
    assert result_ITM_refl == target_ITM_refl

    # Impossible path because of via node.
    with pytest.raises(NetworkXNoPath):
        fabry_perot.path(
            fabry_perot.L0.p1.o, fabry_perot.ITM.p1.o, fabry_perot.ETM.p2.i
        )


def test_fullpath_simple_michelson(simple_michelson):
    """Test path finding for a simple Michelson interferometer (no arm cavities)."""
    # Trace to end of XARM.
    result_XARM = simple_michelson.path(
        simple_michelson.L0.p1.o, simple_michelson.MX.p2.o
    ).data
    target_XARM = [
        (simple_michelson.L0.p1.o, simple_michelson.L0.p1.o.space),
        (simple_michelson.BS.p1.i, simple_michelson.BS),
        (simple_michelson.BS.p3.o, simple_michelson.BS.p3.o.space),
        (simple_michelson.MX.p1.i, simple_michelson.MX),
        (simple_michelson.MX.p2.o, None),
    ]
    assert result_XARM == target_XARM

    # Trace to end of YARM.
    result_YARM = simple_michelson.path(
        simple_michelson.L0.p1.o, simple_michelson.MY.p2.o
    ).data
    target_YARM = [
        (simple_michelson.L0.p1.o, simple_michelson.L0.p1.o.space),
        (simple_michelson.BS.p1.i, simple_michelson.BS),
        (simple_michelson.BS.p2.o, simple_michelson.BS.p2.o.space),
        (simple_michelson.MY.p1.i, simple_michelson.MY),
        (simple_michelson.MY.p2.o, None),
    ]
    assert result_YARM == target_YARM

    # Impossible path through beam splitter.
    with pytest.raises(NetworkXNoPath):
        simple_michelson.path(simple_michelson.MX.p1.o, simple_michelson.MY.p1.i)


def test_fullpath_simple_michelson_with_via_nodes(simple_michelson):
    """Test path finding for a simple Michelson interferometer using specified nodes."""
    # Trace to reflection port of IFO via YARM.
    result_IFO_refl = simple_michelson.path(
        simple_michelson.L0.p1.o,
        simple_michelson.BS.p1.o,
        via_node=simple_michelson.MY.p1.o,
    ).data
    target_IFO_refl = [
        (simple_michelson.L0.p1.o, simple_michelson.L0.p1.o.space),
        (simple_michelson.BS.p1.i, simple_michelson.BS),
        (simple_michelson.BS.p2.o, simple_michelson.BS.p2.o.space),
        (simple_michelson.MY.p1.i, simple_michelson.MY),
        (simple_michelson.MY.p1.o, simple_michelson.MY.p1.o.space),
        (simple_michelson.BS.p2.i, simple_michelson.BS),
        (simple_michelson.BS.p1.o, simple_michelson.BS.p1.o.space),
    ]
    assert result_IFO_refl == target_IFO_refl

    # Trace to AS port of IFO via XARM.
    result_AS = simple_michelson.path(
        simple_michelson.L0.p1.o,
        simple_michelson.BS.p4.o,
        via_node=simple_michelson.MX.p1.i,
    ).data
    target_AS = [
        (simple_michelson.L0.p1.o, simple_michelson.L0.p1.o.space),
        (simple_michelson.BS.p1.i, simple_michelson.BS),
        (simple_michelson.BS.p3.o, simple_michelson.BS.p3.o.space),
        (simple_michelson.MX.p1.i, simple_michelson.MX),
        (simple_michelson.MX.p1.o, simple_michelson.MX.p1.o.space),
        (simple_michelson.BS.p3.i, simple_michelson.BS),
        (simple_michelson.BS.p4.o, None),
    ]
    assert result_AS == target_AS


def test_issue_388():
    """Issue is because of borrowed space ports being used and converted into the owner
    nodes in this case lens1 or l1, and being the wrong direction."""
    import finesse

    model = finesse.Model()
    model.parse(
        """
        l l1 P=10
        gauss g1 l1.p1.o w0=1.25e-3 z=-14e-3
        s s1 l1.p1 len1.p1 L=0.097
        lens len1  f=-76.16e-3
        """
    )

    path = model.path("l1.p1", "s1.p2")

    assert path.nodes[0] is model.l1.p1.o
    assert path.nodes[1] is model.s1.p2.o  # Issue 388

    model = finesse.Model()
    model.parse(
        """
        l l1 P=10
        gauss g1 l1.p1.o w0=1.25e-3 z=-14e-3
        s s1 l1.p1 len1.p1 L=0.097
        lens len1  f=-76.16e-3
        """
    )

    path = model.path("s1.p2", "l1.p1")

    assert path.nodes[0] is model.s1.p2.i
    assert path.nodes[1] is model.l1.p1.i



"""CLI syntax command tests."""

import pytest
from finesse.__main__ import syntax
from finesse.script.spec import KATSPEC
from testutils.cli import sanitized_output


@pytest.mark.parametrize("directive", KATSPEC.directives)
@pytest.mark.parametrize("exact", (False, True))
def test_directive(cli, directive, exact):
    """Search for the specified directive."""
    pieces = [directive]
    if exact:
        pieces.append("--exact")

    cli_result = cli.invoke(syntax, pieces)
    output = sanitized_output(cli_result)
    assert cli_result.exit_code == 0
    assert directive in output



"""Finesse-specific warnings."""


class FinesseWarning(Warning):
    """A generic warning thrown by Finesse."""


class ModelParameterSettingWarning(FinesseWarning):
    """An elements parameter is not using its prefer setter method."""


class CavityUnstableWarning(FinesseWarning):
    """A cavity geometry has become unstable and its eigenmode is no longer
    calculable."""


class KeywordUsedWarning(FinesseWarning):
    """A KatScript keyword was used as a name for some element in the model."""


class InvalidSweepVariableWarning(FinesseWarning):
    """An invalid variable value was used during a sweep action.

    Can occur when trying to sweep relative to an infinite value, or using a NaN.
    """


class UnreasonableComponentValueWarning(FinesseWarning):
    """A not-wrong-but-probably-unreasonable component value was set."""



"""The beam tracing library of Finesse.

This sub-package includes the core data structures and algorithms used for structuring
and performing beam traces (:mod:`.tree` and :mod:`.forest`, for developer reference
mostly), and tools for executing propagations of beams outside of the context of a
simulation (:mod:`.tools`).
"""



"""Misc utility unit tests."""

from hypothesis import given
from hypothesis.strategies import from_regex
from finesse.utilities.misc import check_name


@given(name=from_regex("^[a-zA-Z_][a-zA-Z0-9_]*$", fullmatch=True))
def test_check_name(name):
    """Test that check name is valid for various inputs."""
    check_name(name)



from finesse.simulations.workspace cimport ABCDWorkspace
from finesse.cymath.cmatrix cimport SubCCSView
from finesse.frequency cimport frequency_info_t

from libc.stdlib cimport calloc, free
from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF

cdef class FillFuncWrapper:
    """
    Helper class for wrapping a C fill function that
    can be referenced from Python by objects. This
    allows a direct C call to the function from other
    cdef functions.

    Examples
    --------
    Create a C function then wrap it using this class:

    >>> cdef void c_fill(ConnectorWorkspace ptr_ws) noexcept:
    >>>    cdef MirrorWorkspace ws = <MirrorWorkspace>ptr_ws
    >>>    ...
    >>>
    >>> fill = FillFuncWrapper.make_from_ptr(c_fill)
    """
    def __cinit__(self):
       self.func = NULL

    @staticmethod
    cdef FillFuncWrapper make_from_ptr(fptr_c_fill f) :
        cdef FillFuncWrapper out = FillFuncWrapper()
        out.func = f
        return out


class Connections:
    """This is a container object for storing the connection submatrices defined by
    a component. This is used in the default case where no optimised C class is provided."""
    pass


class NoiseSources:
    """This is a container object for storing the noise submatrices defined by
    a component. This is used in the default case where no optimised C class is provided."""
    pass


cdef class ConnectorMatrixSimulationInfo:
    def __init__(self, object connections=None, object noise_sources=None):
        self.fn_rhs_c = None
        self.fn_rhs_py = None
        self.fn_quantum_noise_c = None
        self.fn_quantum_noise_py = None
        self.fn_quantum_noise_input_c = None
        self.fn_quantum_noise_input_py = None
        self.connections = connections or Connections()
        self.connection_settings = {}
        self.noise_sources = noise_sources or NoiseSources()

    def __dealloc__(self):
        for i in range(self.matrix_fills.size):
            if self.matrix_fills.infos[i].fn_py != NULL:
                Py_XDECREF(self.matrix_fills.infos[i].fn_py)
                self.matrix_fills.infos[i].fn_py = NULL

    def add_fill_function(self, callback, bint refill):
        """ This adds a callback function that will be used by the model to
        fill the matrix elements. This can either be a Python function or cdef
        wrapped with `FillFuncWrapper`.

        Parameters
        ----------
        callback : Callable or FillFuncWrapper
            Callback for fill function
        refill : boolean
            Flags that this fill function will need to be called multiple times
            during simulation to refill the simulation matrix
        """
        cdef Py_ssize_t MAX = sizeof(self.matrix_fills.infos)//sizeof(fill_info_t)

        if self.matrix_fills.size >= MAX:
            raise IndexError(f"Reached maximum number ({MAX}) of fill functions")

        if type(callback) is FillFuncWrapper:
            self.matrix_fills.infos[self.matrix_fills.size].fn_c = (<FillFuncWrapper>callback).func
            self.matrix_fills.infos[self.matrix_fills.size].fn_py = NULL
        elif callable(callback):
            self.matrix_fills.infos[self.matrix_fills.size].fn_py = <PyObject*> callback
            Py_XINCREF(self.matrix_fills.infos[self.matrix_fills.size].fn_py)
        else:
            raise ValueError(f"Callback {callback} should be callable or a FillFuncWrapper")

        self.matrix_fills.infos[self.matrix_fills.size].refill = refill
        self.matrix_fills.size += 1
        self.matrix_fills.num_refills += int(refill)
        self.callback_flag = <ConnectorCallbacks> (self.callback_flag | ConnectorCallbacks.FILL_MATRIX)

    def set_fill_rhs_fn(self, callback):
        """ This sets the callback function that will be used by the model to
        fill the RHS vector elements. This can either be a Python function or cdef
        wrapped with `FillFuncWrapper`.
        """
        if type(callback) is FillFuncWrapper:
            self.fn_rhs_c = callback
        elif callable(callback):
            self.fn_rhs_py = callback
        else:
            raise ValueError(f"Callback {callback} should be callable or a FillFuncWrapper")

        self.callback_flag = <ConnectorCallbacks> (self.callback_flag | ConnectorCallbacks.FILL_RHS)

    def set_fill_noise_function(self, noise_type, callback):
        """ This sets the callback function that will be used by the model to
        fill the noise matrix elements. This can either be a Python
        function, which accepts a
        """
        from finesse.components.general import NoiseType

        if noise_type == NoiseType.QUANTUM:
            if type(callback) is FillFuncWrapper:
                self.fn_quantum_noise_c = callback
            elif callable(callback):
                self.fn_quantum_noise_py = callback
            else:
                raise ValueError(f"Callback {callback} should be callable or a FillFuncWrapper")
        else:
            raise ValueError(f"Unsupported noise type {noise_type}")

        self.callback_flag = <ConnectorCallbacks> (self.callback_flag | ConnectorCallbacks.FILL_NOISE)

    def set_fill_quantum_noise_input_function(self, callback):
        """ This sets the callback function that will be used by the model to
        fill the quantum noise matrix elements due to input nodes.
        """
        if type(callback) is FillFuncWrapper:
            self.fn_quantum_noise_input_c = callback
        elif callable(callback):
            self.fn_quantum_noise_input_py = callback
        else:
            raise ValueError(f"Callback {callback} should be callable or a FillFuncWrapper")

        self.callback_flag = <ConnectorCallbacks> (self.callback_flag | ConnectorCallbacks.FILL_INPUT_NOISE)


cdef class ConnectorWorkspace(ABCDWorkspace):
    """
    This workspace represents the basic container for storing details
    for modelling Connector elements - those which form edges and nodes
    in a model.
    """
    def __init__(
        self,
        object owner,
        object sim,
        object carrier_connections=None,
        object signal_connections=None,
        object values=None,
        object noise_sources=None
    ):
        super().__init__(sim, owner, values=values)
        self.carrier = ConnectorMatrixSimulationInfo(carrier_connections)
        self.signal = ConnectorMatrixSimulationInfo(signal_connections, noise_sources)

        if sim.signal:
            self.setup_quantum_noise()

    cdef setup_quantum_noise(self) :
        cdef:
            int i
            int num_input_nodes
            int num_output_nodes
        from finesse.components.space import Space

        input_nodes = []
        output_nodes = []
        # Spaces don't own any nodes, can't be unconnected, and can't have mismatches across them
        if not isinstance(self.owner, Space):
            for node in self.owner.optical_nodes:
                if node.is_input:
                    # Only add input nodes if they are unconnected
                    # or known as an open port where vacuum noise
                    # gets injected in
                    if not node.port.is_connected:
                        input_nodes.append(node)
                else:
                    output_nodes.append(node)
        num_input_nodes = len(input_nodes)
        num_output_nodes = len(output_nodes)

        self.input_noise = NoiseInfo(num_input_nodes, self.sim.signal.optical_frequencies.size)
        self.output_noise = NoiseInfo(num_output_nodes, self.sim.signal.optical_frequencies.size)

        for i in range(num_input_nodes):
            node = input_nodes[i]
            self.input_noise.node_info[i].idx = self.sim.signal.node_id(node)

        for i in range(num_output_nodes):
            node = output_nodes[i]
            self.output_noise.node_info[i].idx = self.sim.signal.node_id(node)

        self.signal.set_fill_quantum_noise_input_function(optical_quantum_noise_plane_wave)


cdef class NoiseInfo:
    def __init__(self, int num_nodes, int num_frequencies):
        self.num_nodes = num_nodes
        self.nodes = SubCCSView2DArray(num_nodes, num_frequencies)
        self.node_info = <node_noise_info *>calloc(num_nodes, sizeof(node_noise_info))
        if not self.node_info:
            raise MemoryError()
        self.ptrs = self.nodes.views

    def __del__(self):
        if self.node_info:
            free(self.node_info)


optical_quantum_noise_plane_wave = FillFuncWrapper.make_from_ptr(c_optical_quantum_noise_plane_wave)
cdef c_optical_quantum_noise_plane_wave(ConnectorWorkspace ws) :
    cdef:
        int i, j
        double qn
        frequency_info_t *freq

    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        qn  = ws.sim.model_settings.UNIT_VACUUM / 2 * (1 + freq.f_car[0] / ws.sim.model_settings.f0)
        for j in range(ws.input_noise.num_nodes):
            (<SubCCSView>ws.input_noise.ptrs[j][i]).fill_za(qn)



import numpy as np
import scipy.signal as sig

import finesse.components as fc
import finesse.analysis.actions as fa


def test_changing_parameters(model):
    """Tests that masses and moments of inertia can change during a simulation and give
    the correct plants."""
    model.add(fc.Mirror("ETM", T=0, L=0))
    model.add(fc.Mirror("ITM", T=0.014, L=0))
    model.add(fc.FreeMass("ETM_sus", model.ETM.mech))
    model.add(fc.Pendulum("ITM_sus", model.ITM.mech))
    model.fsig.f = F0_Hz = 2

    inertia = 10
    Q = 80
    F_Hz = 5
    dofs = ["z", "pitch", "yaw"]
    props = ["mass", "I_pitch", "I_yaw"]
    for dof in dofs:
        model.set(f"ITM_sus.f{dof}", F0_Hz)
        model.set(f"ITM_sus.Q{dof}", Q)

    to = [f"{tm}_sus.mech.{dof}" for tm in ["ETM", "ITM"] for dof in dofs]
    fr = [f"{tm}_sus.mech.F_{dof}" for tm in ["ETM", "ITM"] for dof in dofs]
    change = {f"{tm}_sus.{p}": inertia for tm in ["ETM", "ITM"] for p in props}

    sol = model.run(
        fa.Series(
            fa.FrequencyResponse([F_Hz], fr, to, name="fresp1"),
            fa.Change(change),
            fa.FrequencyResponse([F_Hz], fr, to, name="fresp2"),
        )
    )

    assert np.all(sol["fresp1"].out == 0)
    free_mass_sus = -1 / (inertia * (2 * np.pi * F_Hz) ** 2)
    pendulum_sus = (
        1 / (inertia * (2 * np.pi) ** 2) / (F0_Hz**2 - F_Hz**2 + 1j * F_Hz * F0_Hz / Q)
    )
    np.testing.assert_allclose(np.diag(sol["fresp2"].out[0, :3, :3]), free_mass_sus)
    np.testing.assert_allclose(np.diag(sol["fresp2"].out[0, 3:, 3:]), pendulum_sus)


def test_SuspensionZPK(model):
    """Tests that SuspensionZPK can correctly model cross-couplings."""
    QQ = 0.01 + 1j
    ps = -np.array([1, 10]) * QQ * 2 * np.pi
    ps = np.concatenate((ps, ps.conjugate()))
    zs = -np.array([5 * QQ, 5 * QQ.conjugate()]) * 2 * np.pi
    zpk_plant = {}
    zpk_plant["z", "F_z"] = (zs, ps, 1)
    zpk_plant["pitch", "F_z"] = ([0], ps, 8)
    model.fsig.f = 1
    model.add(fc.Mirror("ETM", T=0, L=0))
    model.add(fc.SuspensionZPK("ETM_sus", model.ETM.mech, zpk_plant))
    sus = "ETM_sus"
    fr = [f"{sus}.mech.F_z"]
    to = [f"{sus}.mech.{dof}" for dof in ["z", "pitch"]]
    F_Hz = np.geomspace(0.1, 100, 100)
    sol = model.run(fa.FrequencyResponse(F_Hz, fr, to))
    np.testing.assert_allclose(
        sol.out[..., 0, 0],
        sig.freqs_zpk(*zpk_plant["z", "F_z"], 2 * np.pi * F_Hz)[1],
    )
    np.testing.assert_allclose(
        sol.out[..., 1, 0],
        sig.freqs_zpk(*zpk_plant["pitch", "F_z"], 2 * np.pi * F_Hz)[1],
    )



from pathlib import Path

import pytest

from scripts.extract_extra_requirements import extract_reqs, yield_reqs

ROOT = Path(__file__).parent.parent.parent


@pytest.mark.parametrize(
    "toml_content",
    (
        """
[project.optional-dependencies]
inplacebuild = [
    "foo >= 45",
    "bar >= 8.0.3",
    "baz",
]
""",
        """
[project.optional-dependencies]
inplacebuild = [
    "foo >= 45",
    "bar >= 8.0.3",
    "baz",
]
foo = ["foo"]
""",
        """
[project.optional-dependencies]
inplacebuild = [
    "foo >= 45",
    "bar >= 8.0.3",
    "baz",
]
foo = ["foo"]

[new."section]
""",
    ),
)
def test_extract_reqs(toml_content, tmp_path):
    toml_path = tmp_path / "pyproject.toml"
    with open(toml_path, "w") as f:
        f.write(toml_content)
    assert set(yield_reqs(extract_reqs(toml_path, "inplacebuild"))) == {
        "foo >= 45",
        "bar >= 8.0.3",
        "baz",
    }



"""The ``components`` module contains all the component type of an interferometer
configuration including the general objects required to connect them and register node
connections.

These include not only optical components such as mirrors and lasers but also electrical
and mechanical component types found in physical interferometers.
"""

# Note when adding in new elements here you can get circular imports if
# you put them too high up this list.
from .general import Connector, FrequencyGenerator, LocalDegreeOfFreedom, Variable
from .node import Node, NodeType, NodeDirection, Port
from .surface import Surface
from .beamsplitter import Beamsplitter
from .cavity import Cavity
from .gauss import Gauss
from .directional_beamsplitter import DirectionalBeamsplitter
from .optical_bandpass import OpticalBandpassFilter
from .isolator import Isolator
from .laser import Laser
from .lens import Lens, AstigmaticLens
from .mirror import Mirror
from .modulator import Modulator
from .nothing import Nothing
from .frequency_loss import FrequencyLoss
from .readout import ReadoutDC, ReadoutRF
from .signal import SignalGenerator
from .space import Space
from .squeezer import Squeezer
from .wire import Wire
from .electronics import (
    Amplifier,
    Filter,
    ZPKFilter,
    ButterFilter,
    Cheby1Filter,
    TestPoint,
)
from .telescope import Telescope
from .dof import DegreeOfFreedom
from .mechanical import SuspensionTFPlant, FreeMass, Pendulum, SuspensionZPK

__all__ = (
    "LocalDegreeOfFreedom",
    "Connector",
    "Variable",
    "FrequencyGenerator",
    "Surface",
    "Node",
    "NodeType",
    "NodeDirection",
    "Port",
    "Beamsplitter",
    "Cavity",
    "Gauss",
    "DirectionalBeamsplitter",
    "OpticalBandpassFilter",
    "Isolator",
    "Laser",
    "Lens",
    "AstigmaticLens",
    "Mirror",
    "Modulator",
    "Nothing",
    "ReadoutDC",
    "ReadoutRF",
    "SignalGenerator",
    "Space",
    "Squeezer",
    "Wire",
    "Amplifier",
    "Filter",
    "ZPKFilter",
    "ButterFilter",
    "Cheby1Filter",
    "DegreeOfFreedom",
    "SuspensionTFPlant",
    "FreeMass",
    "Pendulum",
    "FrequencyLoss",
    "SuspensionZPK",
    "TestPoint",
    "Telescope",
)



"""Test cases for selecting modes to include in a model."""

import numpy as np
import pytest
from hypothesis import assume, given
from hypothesis.strategies import complex_numbers, composite, floats, integers, one_of

from finesse import Model

HGs_order_2 = [[0, 0], [0, 1], [2, 0], [1, 0], [0, 2], [1, 1]]
HGs_order_1 = [[0, 0], [0, 1], [1, 0]]
HGs_order_4_even = [[0, 0], [2, 0], [0, 2], [4, 0], [2, 2], [0, 4]]
HGs_order_3_odd = [[0, 0], [1, 0], [0, 1], [1, 1], [3, 0], [0, 3]]
HGs_order_5_x = [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0], [5, 0]]
HGs_order_4_y = [[0, 0], [0, 1], [0, 2], [0, 3], [0, 4]]


def test_maxtem_off(model):
    """Test HOMs when maxtem set to off."""
    model.switch_off_homs()
    assert np.all(model.homs == [[0, 0]])


def test_maxtem_zero(model):
    """Test HOMs when maxtem set to 0."""
    model.modes(maxtem=0)
    assert np.all(model.homs == [[0, 0]])


def test_maxtem_on(model):
    """Test HOMs when maxtem set to 2."""
    model.modes(maxtem=2)
    assert HGs_order_2 in model.homs


def test_maxtem_on_increased(model):
    """Test HOMs when maxtem set higher."""
    model.modes(maxtem=1)
    assert HGs_order_1 in model.homs

    model.modes(maxtem=2)
    assert HGs_order_2 in model.homs


def test_maxtem_on_decreased(model):
    """Test HOMs when maxtem set lower."""
    model.modes(maxtem=2)
    assert HGs_order_2 in model.homs

    model.modes(maxtem=1)
    assert HGs_order_1 in model.homs


def test_maxtem_on_to_off(model):
    """Test HOMs when maxtem set to 1 then off."""
    model.modes(maxtem=1)
    assert HGs_order_1 in model.homs

    model.modes("off")
    assert np.all(model.homs == [[0, 0]])


def test_even_modes_specify_maxtem(model):
    """Test HOMs when modes set to even."""
    model.modes("even", 4)

    assert HGs_order_4_even in model.homs


def test_odd_modes_specify_maxtem(model):
    """Test HOMs when modes set to odd."""
    model.modes("odd", 3)

    assert HGs_order_3_odd in model.homs


def test_tangential_modes_specify_maxtem(model):
    """Test HOMs when modes set to 5 for x-direction."""
    model.modes("x", 5)

    assert HGs_order_5_x in model.homs


def test_sagittal_modes_specify_maxtem(model):
    """Test HOMs when modes set to 4 for y-direction."""
    model.modes("y", 4)

    assert HGs_order_4_y in model.homs


def test_insert_single_mode(model):
    """Test HOMs when maxtem set to 3 and modes set to x-direction, with extra mode
    included."""
    model.modes(maxtem=3)
    model.modes("x", 3)
    model.include_modes("11")

    assert [[0, 0], [1, 0], [1, 1], [2, 0], [3, 0]] in model.homs


def test_insert_multiple_modes(model):
    """Test HOMs when maxtem set to even 4 modes, with extra mode included."""
    model.modes("even", 4)
    model.include_modes(["11", "32", "50"])

    assert [
        [0, 0],
        [0, 2],
        [0, 4],
        [1, 1],
        [2, 0],
        [2, 2],
        [3, 2],
        [4, 0],
        [5, 0],
    ] in model.homs


def test_negative_maxtem_is_invalid(model):
    """Test that maxtem cannot be negative."""
    with pytest.raises(ValueError):
        model.modes(maxtem=-2)
    with pytest.raises(ValueError):
        model.modes(maxtem=-314)

    assert np.all(model.homs == [[0, 0]])


def test_non_integer_maxtem_is_invalid(model):
    """Test that maxtem cannot be floats."""
    with pytest.raises(ValueError):
        model.modes(maxtem=3.4)
    with pytest.raises(ValueError):
        model.modes(maxtem=-3.4)

    assert np.all(model.homs == [[0, 0]])


def test_unique_mode_indices(model):
    """Test HOMs when manually selected modes."""

    with pytest.raises(ValueError):
        model.modes(["00", "11", "22", "11", "00"])

    assert np.all(model.homs == [[0, 0]])


@pytest.mark.xfail(reason="See NOTE in Model.modes")
def test_warns_when_enabling(caplog):
    """Test that a warning is emitted when maxtem = 0

    Test that a warning is emitted when
    the model is automatically swicthed
    to modal, but no higher order modes
    are set.

    Also checks that the message is only
    displayed once.
    """
    import logging

    import finesse

    ifo = finesse.Model()
    ifo.parse(
        """

    l LaserIn P=1

    s s0 LaserIn.p1 ITM.p1 L=1

    m ITM R=0.99 T=0.01 Rc=-1
    s sCAV ITM.p2 ETM.p1 L=1.5
    m ETM R=0.99 T=0.01 Rc=1
    """
    )

    assert not ifo.is_modal
    with caplog.at_level(logging.WARNING):
        ifo.parse(
            """
        cavity fab1 ITM.p2 via=ETM.p1.i
        """
        )
        assert "enabled with only HG00" in caplog.text

    assert ifo.is_modal

    # Check that this is only displayed once
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        ifo.parse(
            """
        s s02 ETM.p2 ITM2.p1 L=1
        m ITM2 R=0.99 T=0.01 Rc=-1
        s sCAV2 ITM2.p2 ETM2.p1 L=1.5
        m ETM2 R=0.99 T=0.01 Rc=1
        cavity fab2 ITM2.p2 via=ETM2.p1.i
        """
        )
        assert "enabled with only HG00" not in caplog.text


@composite
def non_whole_floats(draw):
    value = draw(floats())
    if value.is_integer():
        assume(False)
    return value


@given(maxtem=integers(max_value=-1))
def test_negative_integer_maxtem_is_invalid(maxtem):
    """Test that maxtem cannot be negative integer."""
    model = Model()  # Hypothesis can't use the fixture.
    with pytest.raises(ValueError):
        model.modes(maxtem=maxtem)


@given(maxtem=one_of(non_whole_floats(), complex_numbers()))
def test_non_integer_maxtem_is_invalid_fuzzing(maxtem):
    """Test that maxtem cannot be non-integer."""
    model = Model()  # Hypothesis can't use the fixture.
    with pytest.raises((ValueError, TypeError)):
        model.modes(maxtem=maxtem)



"""Test the sequence number defined for each parse call.

A sequence number is created for each parse call, and parsed items are joined to the
root node in the syntax graph by an edge with order attribute set to this number. This
allows the model elements to be later re-sorted into script order even though the parser
does not build elements into the model in that order.
"""

import pytest
from testutils.diff import assert_graph_matches_def


# Override default element to take two arguments.
@pytest.fixture
def fake_element_cls(fake_element_cls):
    class FakeElement(fake_element_cls):
        def __init__(self, name, a, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

    return FakeElement


@pytest.fixture
def spec(spec, fake_element_adapter_factory, fake_element_cls):
    spec.register_element(fake_element_adapter_factory(fake_element_cls))
    return spec


def test_multiple_parses_have_different_sequences(compiler, model):
    """Test that parsing KatScript into the same model multiple times results in
    different branches in the syntax graph."""
    # Graph after first parse.
    refgraph1 = """
        graph [
            directed 1
            node [
                id 0
                label "kat"
                type "ROOT"
                extra_tokens "[]"
            ]
            node [
                id 1
                label "kat.0"
                type "ELEMENT"
                token "'fake_element'"
                name_token "'myelement1'"
                extra_tokens "[' ', ' ']"
            ]
            node [
                id 2
                label "kat.0.0"
                type "VALUE"
                token "'1'"
                key_token "'a'"
                equals_token "'='"
            ]
            edge [
                source 1
                target 0
                type "ARGUMENT"
                order "0"
            ]
            edge [
                source 2
                target 1
                type "ARGUMENT"
                order "0"
            ]
        ]
        """
    compiler.compile("fake_element myelement1 a=1", model=model)
    assert_graph_matches_def(model.syntax_graph, refgraph1)

    # Graph after second parse.
    refgraph2 = """
        graph [
            directed 1
            node [
                id 0
                label "kat"
                type "ROOT"
                extra_tokens "[]"
            ]
            node [
                id 1
                label "kat.0"
                type "ELEMENT"
                token "'fake_element'"
                name_token "'myelement1'"
                extra_tokens "[' ', ' ']"
            ]
            node [
                id 2
                label "kat.0.0"
                type "VALUE"
                token "'1'"
                key_token "'a'"
                equals_token "'='"
            ]
            node [
                id 3
                label "kat.1"
                type "ELEMENT"
                token "'fake_element'"
                name_token "'myelement2'"
                extra_tokens "[' ', ' ', ' ']"
            ]
            node [
                id 4
                label "kat.1.0"
                type "VALUE"
                token "'2'"
                key_token "'a'"
                equals_token "'='"
            ]
            node [
                id 5
                label "kat.1.1"
                type "VALUE"
                token "'3'"
                key_token "'b'"
                equals_token "'='"
            ]
            edge [
                source 1
                target 0
                type "ARGUMENT"
                order "0"
            ]
            edge [
                source 2
                target 1
                type "ARGUMENT"
                order "0"
            ]
            edge [
                source 3
                target 0
                type "ARGUMENT"
                order "1"
            ]
            edge [
                source 4
                target 3
                type "ARGUMENT"
                order "0"
            ]
            edge [
                source 5
                target 3
                type "ARGUMENT"
                order "1"
            ]
        ]
        """
    compiler.compile("fake_element myelement2 a=2 b=3", model=model)
    assert_graph_matches_def(model.syntax_graph, refgraph2)



import pytest
import finesse
import numpy as np
from scipy.special import jv


@pytest.fixture(scope="module")
def modulator_model_out():
    base = finesse.Model()
    base.parse(
        """
        l L0 P=1

        s s0 L0.p1 EOM1.p1

        mod EOM1 f=9M midx=0.1 order=1 mod_type=pm

        pd2 PD2 EOM1.p2.o 9M 0 10
        fsig(10)
        sgen sig L0.amp.i 1 0

        ad icu  EOM1.p1.i fsig.f
        ad ic   EOM1.p1.i 0
        ad icl  EOM1.p1.i -fsig.f

        ad am9u EOM1.p2.o -EOM1.f+fsig.f
        ad am9  EOM1.p2.o -EOM1.f
        ad am9l EOM1.p2.o -EOM1.f-fsig.f

        ad ap9u EOM1.p2.o EOM1.f+fsig.f
        ad ap9  EOM1.p2.o EOM1.f
        ad ap9l EOM1.p2.o EOM1.f-fsig.f

        ad acu  EOM1.p2.o fsig.f
        ad ac   EOM1.p2.o 0
        ad acl  EOM1.p2.o -fsig.f
        """
    )

    return base.run()


def test_carrier_signal_coupling(modulator_model_out):
    sol = modulator_model_out
    fmod = lambda k: ((1j) ** k * jv(k, 0.1))

    assert np.allclose(sol["ac"], fmod(0) * sol["ic"])
    assert np.allclose(sol["acl"], fmod(0) * sol["icl"])
    assert np.allclose(sol["acu"], fmod(0) * sol["icu"])

    assert np.allclose(sol["ap9"], fmod(1) * sol["ic"])
    assert np.allclose(sol["ap9l"], fmod(1) * sol["icl"])
    assert np.allclose(sol["ap9u"], fmod(1) * sol["icu"])

    assert np.allclose(sol["am9"], fmod(-1) * sol["ic"])
    assert np.allclose(sol["am9l"], fmod(-1) * sol["icl"])
    assert np.allclose(sol["am9u"], fmod(-1) * sol["icu"])



import pytest
from finesse.script.exceptions import KatScriptError
from testutils.data import STRINGS
from testutils.tokens import IMPLICITLINEEND, ENDMARKER, STRING


@pytest.mark.parametrize("string,_a,_b", STRINGS)
def test_strings(tokenizer, string, _a, _b):
    assert list(tokenizer.tokenize(string)) == [
        STRING(1, 1, string),
        IMPLICITLINEEND(1, 1 + len(string)),
        ENDMARKER(2),
    ]


@pytest.mark.parametrize(
    "string",
    (
        '"my \nstring"',
        '"my \n\rstring"',
        '"my \r\nstring"',
        '"\n\nmy\r\n\r \n \r \r\nstr\n\ring\n\r"',
    ),
)
def test_strings_cannot_have_newlines(tokenizer, string):
    """Test that strings cannot contain newline characters."""
    with pytest.raises(KatScriptError, match=r"\""):
        list(tokenizer.tokenize(string))



"""Direct probing of coupling coefficients."""

from finesse.detectors.general import Detector
from finesse.detectors.compute.amplitude import (
    KnmDetectorWorkspace,
    knm_detector_scalar_output,
    knm_detector_mode1_output,
    knm_detector_mode2_output,
    knm_detector_matrix_output,
)

from finesse.cymath.homs import field_index


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class KnmDetector(Detector):
    """Direct probe of coupling coefficients at a component.

    This detector has several "modes" which depend upon the values
    given for the mode indices `n1`, `m1` and `n2`, `m2`. If:

      * all of `n1`, `m1`, `n2`, `m2` are specified then the detector
        will output a single complex coefficient corresponding to the
        coupling from ``(n1, m1) -> (n2, m2)``,
      * just `n1` and `m1` are specified then it will output a vector
        of complex coefficients corresponding to each coupling from
        ``(n1, m1) -> (n, m)`` for each mode ``(n, m)`` in the model,
      * only `n2` and `m2` are specified then it will output a vector
        of complex coefficients corresponding to each coupling from
        ``(n, m) -> (n2, m2)`` for each mode ``(n, m)`` in the model,
      * none of `n1`, `m1`, `n2`, `m2` are specified then the detector
        outputs the full matrix of complex coupling coefficients.

    .. hint::

        When using this detector in "full-matrix" mode (i.e. by not giving
        the values for any of the mode indices), it can be useful to combine
        the output with the :class:`.KnmMatrix` object to obtain a more
        convenient representation of the scattering matrix.

        An example of this is shown below, where the output of a detector
        of this type is wrapped using :meth:`.KnmMatrix.from_buffer`.

        .. jupyter-execute::

            import finesse
            finesse.configure(plotting=True)
            from finesse.knm.matrix import KnmMatrix

            IFO = finesse.Model()
            IFO.parse('''
            l L0 P=1
            link(L0, ITM)
            m ITM R=0.99 T=0.01 Rc=-2090 xbeta=0.3u
            s LARM ITM.p2 ETM.p1 L=4k
            m ETM R=0.99 T=0.01 Rc=2090

            cav ARM ITM.p2
            modes(x, maxtem=6)

            knmd K_itm_r ITM 22
            ''')

            out = IFO.run('noxaxis()')
            # Make a KnmMatrix wrapper around the output from the detector
            k_mat = KnmMatrix.from_buffer(out["K_itm_r"], IFO.homs)
            # Now we can perform operations such as plotting the scattering matrix
            k_mat.plot(cmap="bone");

        See :ref:`arbitrary_scatter_matrices` for some examples on the utility that
        the :class:`.KnmMatrix` object provides.

    Parameters
    ----------
    name : str
        Name of newly created KnmDetector.

    comp : :class:`.Connector`
        A component which can scatter modes.

    coupling : str
        Coupling direction string, e.g. "11" for coupling coefficients on reflection from the front
        surface of a mirror.

    n1, m1, n2, m2: int or None
        From (n1, m1) and To (n2, m2) mode indices of the coupling coefficient(s) to retrieve.

        See above for the options.
    """

    def __init__(self, name, comp, coupling, n1=None, m1=None, n2=None, m2=None):
        Detector.__init__(self, name, label="Coupling coefficient", needs_trace=True)

        self.__comp = comp
        self.__coupling = str(coupling)
        if not self.__coupling.startswith("K"):
            self.__coupling = f"K{self.__coupling}"

        int_or_none = lambda x: None if x is None else int(x)
        self.n1 = int_or_none(n1)
        self.m1 = int_or_none(m1)
        self.n2 = int_or_none(n2)
        self.m2 = int_or_none(m2)

        if any(k < 0 for k in (self.n1, self.m1, self.n2, self.m2) if k is not None):
            raise ValueError("Mode indices cannot be negative.")

        for n, m in (("n1", "m1"), ("n2", "m2")):
            sn = getattr(self, n)
            sm = getattr(self, m)
            if (sn is None and sm is not None) or (sm is None and sn is not None):
                raise ValueError(f"Both of {n}, {m} must either be integers or None.")

        self.__mode1_given = self.n1 is not None
        self.__mode2_given = self.n2 is not None

    def _get_workspace(self, sim):
        # TODO (sjr) Move this block to a separate method which should get called
        #            whenever number of modes in model changes, otherwise
        #            querying self.dtype_shape or self.dtype_size will give
        #            incorrect result until simulation performed. Not a big
        #            problem though, as these are typically never accessed by user.
        if self.__mode1_given and self.__mode2_given:
            shape = ()
        elif self.__mode1_given or self.__mode2_given:
            shape = (sim.model_settings.num_HOMs,)
        else:
            shape = (sim.model_settings.num_HOMs, sim.model_settings.num_HOMs)
        self._update_dtype_shape(shape)

        for n, m in [(self.n1, self.m1), (self.n2, self.m2)]:
            if n is None:
                continue

            index = self._model.mode_index_map.get((n, m))
            if index is None:
                raise ValueError(f"The mode ({n},{m}) is not in the model!")

        ws = KnmDetectorWorkspace(self, sim)
        if self.__mode1_given:
            ws.from_idx = field_index(self.n1, self.m1, sim.model_settings.homs_view)
        if self.__mode2_given:
            ws.to_idx = field_index(self.n2, self.m2, sim.model_settings.homs_view)

        comp_ws = list(filter(lambda x: x.owner == self.__comp, sim.workspaces))
        if not comp_ws:
            raise RuntimeError(
                f"Could not find {self.__comp.name} in simulation workspace"
            )

        ws.knm_matrix = getattr(comp_ws[0], self.__coupling, None)
        if ws.knm_matrix is None:
            raise ValueError(
                f"Connector {self.__comp.name} has no scattering "
                f"matrix coupling {self.__coupling}"
            )

        if self.__mode1_given and self.__mode2_given:
            ws.set_output_fn(knm_detector_scalar_output)
        elif self.__mode1_given:
            ws.set_output_fn(knm_detector_mode1_output)
        elif self.__mode2_given:
            ws.set_output_fn(knm_detector_mode2_output)
        else:
            ws.set_output_fn(knm_detector_matrix_output)

        return ws



import pytest
from finesse.utilities.collections import OrderedSet


def test_init():
    oset = OrderedSet([1, 2, 3])
    assert list(oset) == [1, 2, 3]


def test_add():
    oset = OrderedSet()
    oset.add(1)
    assert list(oset) == [1]
    oset.add(2)
    assert list(oset) == [1, 2]
    oset.add(1)  # Adding duplicate
    assert list(oset) == [1, 2]  # Should remain unchanged


def test_remove():
    oset = OrderedSet([1, 2, 3])
    oset.remove(2)
    assert list(oset) == [1, 3]
    with pytest.raises(KeyError):
        oset.remove(4)  # Trying to remove an element not in the set


def test_difference_update():
    oset = OrderedSet([1, 2, 3, 4])
    oset.difference_update([2, 4, 5])
    assert list(oset) == [1, 3]


def test_update():
    oset = OrderedSet([1, 2, 3])
    oset.update([3, 4, 5])
    assert list(oset) == [1, 2, 3, 4, 5]


def test_clear():
    oset = OrderedSet([1, 2, 3])
    oset.clear()
    assert list(oset) == []


def test_contains():
    oset = OrderedSet([1, 2, 3])
    assert 1 in oset
    assert 4 not in oset


def test_len():
    oset = OrderedSet([1, 2, 3])
    assert len(oset) == 3
    oset.add(4)
    assert len(oset) == 4
    oset.remove(2)
    assert len(oset) == 3


def test_iter():
    oset = OrderedSet([1, 2, 3])
    assert list(iter(oset)) == [1, 2, 3]


def test_repr():
    oset = OrderedSet([1, 2, 3])
    assert repr(oset) == "OrderedSet(1, 2, 3)"


def test_ordering():
    # Test ordering on initialization
    oset = OrderedSet([3, 1, 2])
    assert list(oset) == [3, 1, 2], "Initial ordering failed"

    # Test ordering with add
    oset.add(4)
    assert list(oset) == [3, 1, 2, 4], "Ordering failed after adding new element"

    # Adding an existing element should not change order
    oset.add(2)
    assert list(oset) == [
        3,
        1,
        2,
        4,
    ], "Ordering changed after adding an existing element"

    # Test ordering with update
    oset.update([5, 0, 2])
    assert list(oset) == [3, 1, 2, 4, 5, 0], "Ordering failed after update"

    # Test ordering with difference_update
    oset.difference_update([1, 0])
    assert list(oset) == [3, 2, 4, 5], "Ordering failed after difference_update"


def test_sub():
    oset1 = OrderedSet([1, 2, 3, 4])
    oset2 = OrderedSet([3, 4, 5])
    oset3 = oset1 - oset2
    assert list(oset3) == [1, 2], "Subtraction did not return the expected result"

    oset4 = OrderedSet([1, 2])
    oset5 = oset1 - oset4
    assert list(oset5) == [
        3,
        4,
    ], "Subtraction did not return the expected result when subtracting another subset"


def test_union():
    # Test union of two OrderedSets with some common elements
    oset1 = OrderedSet([1, 2, 3])
    oset2 = OrderedSet([3, 4, 5])
    oset3 = oset1 | oset2
    assert list(oset3) == [1, 2, 3, 4, 5], "Union failed with common elements"

    # Test union with an empty OrderedSet
    oset4 = OrderedSet()
    oset5 = oset1 | oset4
    assert list(oset5) == [
        1,
        2,
        3,
    ], "Union with an empty set should return the original set"

    # Test union where one set is a subset of the other
    oset6 = OrderedSet([1, 2])
    oset7 = oset1 | oset6
    assert list(oset7) == [
        1,
        2,
        3,
    ], "Union failed when one set was a subset of the other"

    # Test union where both sets are identical
    oset8 = oset1 | oset1
    assert list(oset8) == [
        1,
        2,
        3,
    ], "Union of identical sets should return the same set"

    # Test union with no common elements
    oset9 = OrderedSet([6, 7])
    oset10 = oset1 | oset9
    assert list(oset10) == [
        1,
        2,
        3,
        6,
        7,
    ], "Union with no common elements should return all elements from both sets"


def test_intersection():
    # Test intersection of two OrderedSets with some common elements
    oset1 = OrderedSet([1, 2, 3])
    oset2 = OrderedSet([2, 3, 4])
    oset3 = oset1 & oset2
    assert list(oset3) == [2, 3], "Intersection failed with common elements"

    # Test intersection with an empty OrderedSet
    oset4 = OrderedSet()
    oset5 = oset1 & oset4
    assert (
        list(oset5) == []
    ), "Intersection with an empty set should return an empty set"

    # Test intersection where one set is a subset of the other
    oset6 = OrderedSet([1, 2])
    oset7 = oset1 & oset6
    assert list(oset7) == [
        1,
        2,
    ], "Intersection failed when one set was a subset of the other"

    # Test intersection where both sets are identical
    oset8 = oset1 & oset1
    assert list(oset8) == [
        1,
        2,
        3,
    ], "Intersection of identical sets should return the same set"

    # Test intersection with no common elements
    oset9 = OrderedSet([4, 5])
    oset10 = oset1 & oset9
    assert (
        list(oset10) == []
    ), "Intersection with no common elements should return an empty set"


def test_issubset():
    oset1 = OrderedSet([1, 2, 3])
    oset2 = OrderedSet([1, 2, 3, 4, 5])
    oset3 = OrderedSet([2, 3])
    oset4 = OrderedSet([1, 4])

    assert oset1.issubset(oset2) is True, "Set should be a subset of the other"
    assert oset1.issubset(oset3) is False, "Set should not be a subset of the other"
    assert oset3.issubset(oset1) is True, "Set should be a subset of the other"
    assert oset4.issubset(oset1) is False, "Set should not be a subset of the other"


def test_equality():
    oset1 = OrderedSet([1, 2, 3])
    oset2 = OrderedSet([1, 2, 3])
    oset3 = OrderedSet([3, 2, 1])
    oset4 = OrderedSet([1, 2, 4])
    oset5 = OrderedSet()
    native_set = {1, 2, 3}
    native_set_different = {2, 3, 4}

    # Test equality with another OrderedSet
    assert (
        oset1 == oset2
    ), "Sets with the same elements in the same order should be equal"
    assert (
        oset1 != oset3
    ), "Sets with the same elements but different order should not be equal"
    assert oset1 != oset4, "Sets with different elements should not be equal"
    assert oset1 != oset5, "Sets with different elements should not be equal"
    assert oset5 == OrderedSet(), "Empty sets should be equal"

    # Test equality with other collection types
    assert (
        oset1 == native_set
    ), "OrderedSet should be equal to a set with the same elements"
    assert (
        oset1 != native_set_different
    ), "OrderedSet should not be equal to a set with different elements"
    assert oset1 != [1, 2, 3], "OrderedSet should not be equal to a list"
    assert oset1 != (1, 2, 3), "OrderedSet should not be equal to a tuple"
    assert oset1 != "123", "OrderedSet should not be equal to a string"
    assert oset1 != 123, "OrderedSet should not be equal to an integer"



from finesse.detectors.general import Detector
from finesse.detectors.workspace import DetectorWorkspace
from finesse.utilities import OrderedSet


class MathDetectorWorkspace(DetectorWorkspace):
    """MathDetectorWorkspace."""

    def __init__(self, owner, sim):
        from finesse.components.readout import ReadoutDetectorOutput
        from finesse.symbols import Symbol, Constant, Variable

        self.expression = owner.expression
        if not isinstance(self.expression, Symbol):
            self.expression = Constant(self.expression)

        # Get all the detectors in the expression
        self.detectors = [  # will select constants, whose value is a detector
            a
            for a in self.expression.all(
                lambda a: isinstance(a, (Constant, Variable))
                and isinstance(a.value, (Detector, ReadoutDetectorOutput))
            )
        ]
        out_wss = OrderedSet(  # workspaces can be in both lists
            (*sim.readout_workspaces, *sim.detector_workspaces)
        )
        # find the workspaces for the detectors
        self.dws = []
        for det in self.detectors:
            for ws in out_wss:
                if ws.oinfo.name == det.value.name:
                    self.dws.append(ws)
        self.set_output_fn(self.__output)
        needs_carrier = any(_.needs_carrier for _ in self.dws)
        needs_signal = any(_.needs_signal for _ in self.dws)
        needs_noise = any(_.needs_noise for _ in self.dws)
        needs_modal_update = any(_.needs_modal_update for _ in self.dws)
        needs_simulation = (
            needs_carrier or needs_signal or needs_noise or needs_modal_update
        )
        super().__init__(
            owner,
            sim,
            needs_carrier=needs_carrier,
            needs_signal=needs_signal,
            needs_noise=needs_noise,
            needs_modal_update=needs_modal_update,
            needs_simulation=needs_simulation,
        )

    def __output(self, ws):
        subs = {a: ws.get_output() for a, ws in zip(self.detectors, self.dws)}
        return ws.expression.eval(subs=subs)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class MathDetector(Detector):
    """A detector that performs some math operation and outputs the result.

    Parameters
    ----------
    name : str
        Name of detector

    expression : Symbol
        Symbolic expression to evaluate as the detectors output

    dtype : numpy.dtype, optional
        The expected data type when evaluating expression. Defaults to "O".

    dtype_shape : tuple
        The expected array shape when evaluating `expression`. Defaults to an empty tuple.
        This must be an empty tuple if `dtype` is a scalar type or "O".

    unit : str
        The unit of the output for plotting. Defaults to "arb."

    label : str or None
        How to label the axis when plotting this detector. Defaults to `None`.

    Notes
    -----
    If `dtype` does not match the result when `expression` is evaluated the values will be
    silently cast to `dtype`.
    `dtype_shape` is used to create a numpy array in which the results are written, if this
    doesn't match this should raise some Exception when running a simulation, but this is
    not guaranteed.
    If one isn't sure of what type `expression` will produce, one should leave the default "O",
    the generic numpy type `object`, and convert the data after inspecting the result.
    When using the default type, some functions like `Solution.plot` might not be able to
    convert the data to an appropiate type internally, in which case they cannot be used.

    Examples
    --------
    KatScript example:

        l l1 P=1
        pd P l1.p1.o
        fd E l1.p1.o f=l1.f
        bp qx l1.p1.o q x
        modes(maxtem=3)
        gauss g1 l1.p1.o w0=1m z=0

        mathd Y1 P*2
        mathd Y2 P**2
        mathd Y3 cos(1+P**2)
        mathd Y4 E*2
        mathd Y5 qx+1
    """

    def __init__(
        self, name, expression, dtype="O", dtype_shape=tuple(), unit="arb.", label=None
    ):
        super().__init__(name, dtype=dtype)
        self._expression = expression
        self._update_dtype_shape(dtype_shape)
        self._update_unit(unit)
        self._update_label(label)

    @property
    def needs_fields(self):
        return False

    @property
    def needs_trace(self):
        return False

    def set_expression(
        self, expression, dtype="O", dtype_shape=tuple(), unit="arb.", label=None
    ):
        self._expression = expression
        self._update_dtype(dtype)
        self._update_dtype_shape(dtype_shape)
        self._update_unit(unit)
        self._update_label(label)

    @property
    def expression(self):
        return self._expression

    def _get_workspace(self, sim):
        return MathDetectorWorkspace(self, sim)



from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport cexp
from finesse.cymath.math cimport sqrt, radians
from finesse.cymath.gaussbeam cimport bp_gouy, bp_beamsize
from finesse.cymath.cmatrix cimport SubCCSView
from finesse.simulations.base cimport NodeBeamParam
from finesse.frequency cimport frequency_info_t
from finesse.cymath.math cimport sgn
from finesse.simulations.simulation cimport BaseSimulation
from finesse.simulations.homsolver cimport HOMSolver
from finesse.cymath.complex cimport DenseZVector
from finesse.knm.matrix cimport make_unscaled_X_scatter_knm_matrix, make_unscaled_Y_scatter_knm_matrix
from finesse.simulations.workspace cimport ABCDWorkspace

import numpy as np
cimport numpy as np
cimport cython

from cpython.ref cimport PyObject

ctypedef (double*, double*, double*) ptr_tuple_3

cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD

cdef class LaserConnections:
    def __cinit__(self, HOMSolver mtx):
        cdef:
            int Nfo =  mtx.optical_frequencies.size

        # There are no carrier connections at a laser, just signals
        self.SIGPWR_P1o = SubCCSView2DArray(1, Nfo)
        self.SIGAMP_P1o = SubCCSView2DArray(1, Nfo)
        self.SIGFRQ_P1o = SubCCSView2DArray(1, Nfo)
        self.SIGPHS_P1o = SubCCSView2DArray(1, Nfo)
        self.dz_P1o = SubCCSView2DArray(1, Nfo)
        self.dx_P1o = SubCCSView2DArray(1, Nfo)
        self.dy_P1o = SubCCSView2DArray(1, Nfo)
        self.xbeta_P1o = SubCCSView2DArray(1, Nfo)
        self.ybeta_P1o = SubCCSView2DArray(1, Nfo)

        self.ptrs.SIGPWR_P1o = <PyObject***>self.SIGPWR_P1o.views
        self.ptrs.SIGAMP_P1o = <PyObject***>self.SIGAMP_P1o.views
        self.ptrs.SIGFRQ_P1o = <PyObject***>self.SIGFRQ_P1o.views
        self.ptrs.SIGPHS_P1o = <PyObject***>self.SIGPHS_P1o.views
        self.ptrs.dz_P1o = <PyObject***>self.dz_P1o.views
        self.ptrs.dx_P1o = <PyObject***>self.dx_P1o.views
        self.ptrs.dy_P1o = <PyObject***>self.dy_P1o.views
        self.ptrs.xbeta_P1o = <PyObject***>self.xbeta_P1o.views
        self.ptrs.ybeta_P1o = <PyObject***>self.ybeta_P1o.views


cdef class LaserValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_3 ptr = (&self.P, &self.phase, &self.f)
        cdef tuple params = ("P", "phase", "f")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class LaserWorkspace(ConnectorWorkspace):
    def __init__(self, object owner, BaseSimulation sim):
        super().__init__(
                owner,
                sim,
                None,
                LaserConnections(sim.signal) if sim.signal else None,
                LaserValues()
                )
        self.cvalues = self.values
        self.lc = self.signal.connections if sim.signal else None
        self.PIj_2 = PI*0.5j
        # indexes for beam tracing
        self.P1o_id = sim.trace_node_index[owner.p1.o]
        self.K_yaw_sig = make_unscaled_X_scatter_knm_matrix(sim.model_settings.homs_view)
        self.K_pitch_sig = make_unscaled_Y_scatter_knm_matrix(sim.model_settings.homs_view)
        self.hom_vector = np.zeros(sim.model_settings.num_HOMs, dtype=complex)
        self.phase_vector = np.zeros(sim.model_settings.num_HOMs, dtype=complex)


cdef complex_t laser_scalar(LaserWorkspace ws) noexcept:
    """Scalar value for laser output"""
    return sqrt(2 * ws.cvalues.P / ws.sim.model_settings.EPSILON0_C) * cexp(1.0j * radians(ws.cvalues.phase))


cdef void fill_hom_vector(complex_t E, LaserWorkspace ws) noexcept:
    """Fills the workspace HOM vector for this laser"""
    cdef Py_ssize_t i

    if ws.sim.is_modal:
        for i in range(ws.sim.model_settings.num_HOMs):
            ws.hom_vector[i] = E * ws.power_coeffs[i] * ws.phase_vector[i]
    else:
        ws.hom_vector[0] = E # planewave


laser_carrier_fill_rhs = FillFuncWrapper.make_from_ptr(c_laser_carrier_fill_rhs)
cdef object c_laser_carrier_fill_rhs(ConnectorWorkspace cws) :
    r"""
    Fills the right hand side (RHS) vector corresponding to the light source `laser`.

    The field amplitude is set as

    .. math::
        a_{\mathrm{in}} = \sqrt{\frac{2P}{\epsilon_c}}~\exp{\left(i \varphi\right)},

    where :math:`P` is the laser power and :math:`\varphi` is the specified phase of
    the laser.

    Parameters
    ----------

    laser : :class:`.Laser`
        The laser object to fill.

    sim : :class:`.BaseSimulation`
        A handle to the simulation.

    values : dict
        Dictionary of evaluated model parameters.

    fsrc_index : int
        Index of source frequency bin.
    """
    cdef:
        LaserWorkspace ws = <LaserWorkspace>cws
        HOMSolver carrier = ws.sim.carrier
        complex_t Ein = laser_scalar(ws)
        Py_ssize_t i

    if ws.cvalues.signals_only:
        return

    if not ws.sim.is_modal:
        carrier.set_source_fast(
            ws.node_car_id, ws.fsrc_car_idx, 0, Ein, 0
        )
    else:
        fill_hom_vector(Ein, ws)
        for i in range(ws.sim.model_settings.num_HOMs):
            carrier.set_source_fast(
                ws.node_car_id, ws.fsrc_car_idx, i, ws.hom_vector[i], 0
            )


laser_fill_qnoise = FillFuncWrapper.make_from_ptr(c_laser_fill_qnoise)
cdef object c_laser_fill_qnoise(ConnectorWorkspace cws) :
    r"""
    Fills the quantum noise input matrix corresponding to the light source `laser`.
    """
    cdef:
        LaserWorkspace ws = <LaserWorkspace>cws
        PyObject ***noises = ws.output_noise.ptrs
        frequency_info_t *freq

        # Laser quantum noise injection
        complex_t qn

    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        qn = ws.sim.model_settings.UNIT_VACUUM / 2 * (1 + freq.f_car[0] / ws.sim.model_settings.f0)
        (<SubCCSView>noises[0][freq.index]).fill_za(qn)


laser_fill_signal = FillFuncWrapper.make_from_ptr(c_laser_fill_signal)
cdef object c_laser_fill_signal(ConnectorWorkspace cws) :
    cdef:
        LaserWorkspace ws = <LaserWorkspace>cws
        laser_connections conns = <laser_connections>ws.lc.ptrs
        Py_ssize_t i
        double factor = ws.sim.model_settings.EPSILON0_C

        complex_t phs_sig
        complex_t frq_sig

        frequency_info_t *f
        DenseZVector c_p1_o
        NodeBeamParam *q_P1o
        double w
        double k = ws.sim.model_settings.k0

        complex_t Ein

    # Recompute any input field changes
    Ein = laser_scalar(ws)
    fill_hom_vector(Ein, ws)
    # fixed definitions for vector
    c_p1_o.size = ws.sim.model_settings.num_HOMs
    c_p1_o.stride = 1
    c_p1_o.ptr = &ws.hom_vector[0]

    for i in range(2): # Loop over each signal sideband
        f = &ws.sim.signal.optical_frequencies.frequency_info[ws.fcar_sig_sb_idx[i]]
        # TODO ddb - these are all assuming a single electronic frequency here, hence first 0 index
        if conns.SIGAMP_P1o[0][f.index]:
            # m/2 to get 2 * m * cosine power fluctuations
            (<SubCCSView>conns.SIGAMP_P1o[0][f.index]).fill_negative_za_zv(factor * 0.5, &c_p1_o)

        if conns.SIGPWR_P1o[0][f.index]:
            # m/4 to get m * cosine power fluctuations
            (<SubCCSView>conns.SIGPWR_P1o[0][f.index]).fill_negative_za_zv(factor * 0.5 * 0.5, &c_p1_o)

        if conns.SIGFRQ_P1o[0][f.index]:
            frq_sig =  0.5 / ws.sim.model_settings.fsig * sgn(f.audio_order)
            (<SubCCSView>conns.SIGFRQ_P1o[0][f.index]).fill_negative_za_zv(factor * frq_sig, &c_p1_o)

        if conns.SIGPHS_P1o[0][f.index]:
            phs_sig = 1j * 0.5
            (<SubCCSView>conns.SIGPHS_P1o[0][f.index]).fill_negative_za_zv(factor * phs_sig, &c_p1_o)

        if conns.dz_P1o[0][f.index]:
            (<SubCCSView>conns.dz_P1o[0][f.index]).fill_negative_za_zv(0.5j * k * factor, &c_p1_o)

        if conns.dx_P1o[0][f.index] or conns.xbeta_P1o[0][f.index]:
            q_P1o = &ws.sim.trace[ws.P1o_id]
            w = bp_beamsize(&q_P1o.qx)
            if conns.dx_P1o[0][f.index]:
                (<SubCCSView>conns.dx_P1o[0][f.index]).fill_negative_za_zmv(0.5 * k *  factor * w, &ws.K_yaw_sig.mtx, &c_p1_o)
            if conns.xbeta_P1o[0][f.index]:
                (<SubCCSView>conns.xbeta_P1o[0][f.index]).fill_negative_za_zmv(0.5j * k * factor * w, &ws.K_yaw_sig.mtx, &c_p1_o)

        if conns.dy_P1o[0][f.index] or conns.ybeta_P1o[0][f.index]:
            q_P1o = &ws.sim.trace[ws.P1o_id]
            w = bp_beamsize(&q_P1o.qy)
            if conns.dy_P1o[0][f.index]:
                (<SubCCSView>conns.dy_P1o[0][f.index]).fill_negative_za_zmv(0.5 * k * factor * w, &ws.K_pitch_sig.mtx, &c_p1_o)
            if conns.ybeta_P1o[0][f.index]:
                (<SubCCSView>conns.ybeta_P1o[0][f.index]).fill_negative_za_zmv(0.5j * k * factor * w, &ws.K_pitch_sig.mtx, &c_p1_o)


laser_set_gouy = GouyFuncWrapper.make_from_ptr(set_tem_gouy_phases)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef int set_tem_gouy_phases(ABCDWorkspace ws) except -1:
    cdef:
        const NodeBeamParam* q_p1o = &ws.sim.trace[ws.P1o_id]
        double gouy_x
        double gouy_y
        double phase00 = 0.0
        Py_ssize_t i
        int n, m

    gouy_x = bp_gouy(&q_p1o.qx)
    gouy_y = bp_gouy(&q_p1o.qy)

    if ws.sim.model_settings.phase_config.zero_tem00_gouy:
        phase00 = 0.5 * gouy_x + 0.5 * gouy_y

    for i in range(ws.sim.model_settings.num_HOMs):
        if ws.add_gouy_phase:
            n = ws.sim.model_settings.homs_view[i][0]
            m = ws.sim.model_settings.homs_view[i][1]
            ws.phase_vector[i] = cexp(1j*(((n + 0.5) * gouy_x + (m + 0.5) * gouy_y - phase00)))
        else:
            ws.phase_vector[i] = 1

    return 0



"""List available Finesse element validators.

Originally used for #123.
"""

from list_elements import MODEL_ELEMENT_CHILDREN


def model_element_validators(klass):
    """Dict of parameter names to validator callables."""
    # Reverse param dict since it's defined in reverse.
    for pinfo in reversed(klass._param_dict[klass]):
        try:
            validator = klass._validators[klass][pinfo.name]
        except KeyError:
            validator = None

        yield pinfo.name, validator


if __name__ == "__main__":
    for cls in sorted(
        MODEL_ELEMENT_CHILDREN, key=lambda cls: (cls.__module__, cls.__name__)
    ):
        print(f"### `{cls.__module__}.{cls.__name__}`")
        print()

        validators = dict(model_element_validators(cls))

        if validators:
            for parameter, validator in validators.items():
                status = " --- **no validator**" if validator is None else ""
                print(f"* [ ] `{parameter}`{status}")
        else:
            print("*No model parameters.*")

        print()



import pytest

from finesse.analysis.actions.base import Action
from finesse.components.general import Connector
from finesse.components.node import NodeDirection, NodeType
from finesse.model import Model
from finesse.parameter import float_parameter
from finesse.script.adapter import (
    AnalysisDocumenter,
    AnalysisDumper,
    AnalysisFactory,
    AnalysisSetter,
    CommandMethodDocumenter,
    CommandMethodSetter,
    ElementDocumenter,
    ElementDumper,
    ElementFactory,
    ElementSetter,
    ItemAdapter,
)
from finesse.script.spec import KatSpec
from finesse.symbols import FUNCTIONS, OPERATORS


@pytest.fixture
def spec(monkeypatch):
    """Kat spec with no registered directives or language constructs.

    Tests that use this should register what they need each time. This ensures only the
    constructs being tested are present.
    """

    spec = KatSpec()

    ## Clear existing spec.
    # Directives.
    spec.elements = {}
    spec.commands = {}
    spec.analyses = {}
    # Language constructs.
    monkeypatch.setattr(spec, "keywords", set())
    monkeypatch.setattr(spec, "constants", {})
    monkeypatch.setattr(spec, "unary_operators", {})
    monkeypatch.setattr(spec, "binary_operators", {})
    monkeypatch.setattr(spec, "expression_functions", {})

    return spec


@pytest.fixture
def set_spec_constructs(monkeypatch, spec):
    """Register a language construct in the spec.

    This fixture provides a callable that accepts a sequence containing zero or more
    construct name, new value subsequences.

    Examples
    --------
    Set the available keywords and constants to exactly "lin" and "log", and "c" and
    "pi", respectively:

    >>> set_spec_constructs("keywords", ("lin", "log"), "constants", ("c", "pi"))
    """

    constructs = (
        "keywords",
        "constants",
        "unary_operators",
        "binary_operators",
        "expression_functions",
    )

    def _(*args):
        for construct, newval in zip(args[0::2], args[1::2]):
            assert construct in constructs
            monkeypatch.setattr(spec, construct, newval)

    return _


@pytest.fixture
def model():
    class FakeModel(Model):
        """A fake model."""

    return FakeModel()


@pytest.fixture
def fake_adapter():
    return ItemAdapter


@pytest.fixture
def fake_float_parameter():
    return float_parameter


@pytest.fixture
def fake_element_cls():
    """A minimal element.

    Test modules may override this to add more arguments.
    """

    class FakeElement(Connector):
        """A fake element that test modules should override."""

        def __init__(self, name):
            super().__init__(name)

            # Add some ports (required by the parent class).
            self._add_port("p1", NodeType.OPTICAL)
            self.p1._add_node("i", NodeDirection.INPUT)
            self.p1._add_node("o", NodeDirection.OUTPUT)

            self._add_port("p2", NodeType.OPTICAL)
            self.p2._add_node("i", NodeDirection.INPUT)
            self.p2._add_node("o", NodeDirection.OUTPUT)

    return FakeElement


@pytest.fixture
def fake_command_func():
    """A minimal command.

    Test modules may override this to add more arguments.
    """

    def fake_command(model):
        pass

    return fake_command


@pytest.fixture
def fake_analysis_cls():
    """A minimal analysis.

    Test modules may override this to add more arguments.
    """

    # Needs to inherit Action so the generator matches nested analyses correctly.
    class FakeAnalysis(Action):
        def __init__(self):
            pass

        def _requests(self, *args, **kwargs):
            pass

        def _do(self, state):
            pass

    return FakeAnalysis


@pytest.fixture
def fake_element_adapter_factory(fake_adapter):
    def _(
        element_cls,
        full_name="fake_element",
        getter_kwargs=None,
        factory_kwargs=None,
        setter_kwargs=None,
        documenter_kwargs=None,
        **adapter_kwargs,
    ):
        getter_kwargs = {} if getter_kwargs is None else getter_kwargs
        factory_kwargs = {} if factory_kwargs is None else factory_kwargs
        setter_kwargs = {} if setter_kwargs is None else setter_kwargs
        documenter_kwargs = {} if documenter_kwargs is None else documenter_kwargs
        return fake_adapter(
            full_name,
            getter=ElementDumper(item_type=element_cls, **getter_kwargs),
            factory=ElementFactory(item_type=element_cls, **factory_kwargs),
            setter=ElementSetter(item_type=element_cls, **setter_kwargs),
            documenter=ElementDocumenter(item_type=element_cls, **documenter_kwargs),
            **adapter_kwargs,
        )

    return _


@pytest.fixture
def fake_analysis_adapter_factory(fake_adapter):
    def _(
        analysis_cls,
        full_name="fake_analysis",
        getter_kwargs=None,
        factory_kwargs=None,
        setter_kwargs=None,
        documenter_kwargs=None,
        **adapter_kwargs,
    ):
        getter_kwargs = {} if getter_kwargs is None else getter_kwargs
        factory_kwargs = {} if factory_kwargs is None else factory_kwargs
        setter_kwargs = {} if setter_kwargs is None else setter_kwargs
        documenter_kwargs = {} if documenter_kwargs is None else documenter_kwargs
        return fake_adapter(
            full_name,
            getter=AnalysisDumper(item_type=analysis_cls, **getter_kwargs),
            factory=AnalysisFactory(item_type=analysis_cls, **factory_kwargs),
            setter=AnalysisSetter(item_type=analysis_cls, **setter_kwargs),
            documenter=AnalysisDocumenter(item_type=analysis_cls, **documenter_kwargs),
            singular=True,
            **adapter_kwargs,
        )

    return _


@pytest.fixture
def fake_command_adapter_factory(fake_adapter):
    # Assume the passed `command_func` is a function, not method, so we exclude the
    # first argument `model` by default.
    def _(
        command_func,
        full_name="fake_command",
        sig_ignore=("model",),
        setter=None,
        setter_kwargs=None,
        documenter=None,
        documenter_kwargs=None,
        **adapter_kwargs,
    ):
        setter_kwargs = {} if setter_kwargs is None else setter_kwargs
        documenter_kwargs = {} if documenter_kwargs is None else documenter_kwargs
        if setter is None:
            setter = CommandMethodSetter(
                item_type=command_func, sig_ignore=sig_ignore, **setter_kwargs
            )

        if documenter is None:
            documenter = CommandMethodDocumenter(
                item_type=command_func, sig_ignore=sig_ignore, **documenter_kwargs
            )

        return fake_adapter(
            full_name,
            setter=setter,
            documenter=documenter,
            **adapter_kwargs,
        )

    return _


@pytest.fixture
def fake_noop():
    """A fake operation that should not be called. If called, it triggers an error.

    This is useful for use in tests where the operator is not expected to be called.
    """

    def noop(*args, **kwargs):
        raise RuntimeError(
            "This should not have been called, and indicates a test error. "
            "Make or use a Finesse operation if the results of this function are to be "
            "read."
        )

    return noop


@pytest.fixture
def finesse_binop_add():
    return OPERATORS["__add__"]


@pytest.fixture
def finesse_binop_sub():
    return OPERATORS["__sub__"]


@pytest.fixture
def finesse_binop_mul():
    return OPERATORS["__mul__"]


@pytest.fixture
def finesse_binop_div():
    return OPERATORS["__truediv__"]


@pytest.fixture
def finesse_binop_pow():
    return OPERATORS["__pow__"]


@pytest.fixture
def finesse_unop_neg():
    return FUNCTIONS["neg"]



"""Base level Actions utilities and classes."""

# Allow generics in type hints (PEP 585). This can be removed once Finesse requires at
# least Python 3.9.
from __future__ import annotations

import abc
import logging
import time
from collections import defaultdict

import finesse
from finesse.solutions import BaseSolution
from finesse.tree import TreeNode
from finesse.utilities.components import names_to_nodes

LOGGER = logging.getLogger(__name__)


def convert_str_to_parameter(model, attr):
    """Converts names `component.parameter` or `component` to a parameter object. Will
    return default parameter when component name is given.

    Parameters
    ----------
    model : Model
        Model object to look for parameter in
    attr : [str | Parameter]
        String value for the name of an element or a parameters full name. If
        a Parameter is given its full name will be used to grab the equivalent
        parameter in this Model.

    Returns
    -------
    parameter
        The equivalent Parameter object for the attr provided
    """
    if hasattr(attr, "full_name"):
        return model.get(attr.full_name)
    else:
        obj = model.get(attr)

        # If this attr string has no period in it, assume it is an element name
        # and try and get it
        if "." in attr:
            return obj
        else:
            if isinstance(obj, finesse.parameter.Parameter):
                return obj

            if obj.default_parameter_name is None:
                raise ValueError(
                    f"{repr(obj)} does not have a default parameter, please specify one to use"
                )
            return getattr(obj, obj.default_parameter_name)


def request_dict_reduction(A, B):
    dd = defaultdict(list)
    for d in (A, B):
        for key, value in d.items():
            dd[key].extend(value)
    return dd


class AnalysisState(TreeNode):
    def __init__(
        self,
        model,
        name="AnalysisState",
        parent=None,
        simulation_type=None,
        simulation_options=None,
    ):
        super().__init__(f"{name} {model}", parent=parent)
        assert isinstance(model, finesse.model.Model)
        self.__model = model
        self.__simulation_type = simulation_type
        self.__simulation_options = simulation_options
        self.__sim = None
        self.__previous_solution = None
        self.model_finished_with = True
        self.__action_workspaces = {}

    @property
    def model(self):
        return self.__model

    @property
    def action_workspaces(self):
        """Actions can use their id, `id(self)`, to generate a key to store simulation
        specific data and reuse it each time run is called."""
        return self.__action_workspaces

    @property
    def sim(self):
        return self.__sim

    @property
    def previous_solution(self):
        return self.__previous_solution

    def apply(self, action):
        start = time.time_ns()
        sol = action._do(self)
        if sol is not None:
            if not isinstance(sol, BaseSolution):
                raise TypeError(
                    f"Action of type {type(action)} should return a BaseSolution derivative, not {sol}"
                )
            sol.time = (time.time_ns() - start) / 1e9
            self.__previous_solution = sol
        return sol

    def _split(self):
        state = AnalysisState(self.model.deepcopy(), parent=self)
        return state

    def build_model(self, changing_params, extra_input_nodes, extra_output_nodes):
        if not self.model_finished_with:
            raise Exception(
                "Trying to build new model whilst current one is in use. Make sure to call `finished()` on this state if the simulation has been completed."
            )

        if self.model.is_built:
            self.finished()

        LOGGER.info(
            f"Building simulation for model {repr(self.model)}"
            f"Changing parameters = {changing_params}"
        )

        # If we do not have a simulation we need to build one
        for p in changing_params:
            p.is_tunable = True

        self.keep_nodes = tuple(
            names_to_nodes(self.model, extra_input_nodes + extra_output_nodes)
        )

        # Tell node it is being used as some sort of output so it doesn't get removed
        # TODO ddb : could refactor the naming for more generic use instead of detector
        for obj in self.keep_nodes:
            obj.used_in_detector_output.append(self)

        sim_opts = {
            "extra_input_nodes": extra_input_nodes,
            "extra_output_nodes": extra_output_nodes,
        }
        if self.__simulation_options is not None:
            sim_opts.update(self.__simulation_options)

        self.__changing_params = changing_params
        self.__sim = self.model._build(
            simulation_type=self.__simulation_type, simulation_options=sim_opts
        )
        self.__sim.__enter__()
        self.model_finished_with = False

    def finished(self):
        if self.__sim:
            LOGGER.info(
                f"Finishing simulation {repr(self.sim)} for model {repr(self.model)}"
            )
            self.model_finished_with = True
            self.__sim.__exit__(None, None, None)
            self.model.unbuild()
            for p in self.__changing_params:
                p.is_tunable = False
            for obj in self.keep_nodes:
                obj.used_in_detector_output.remove(self)
            self.__sim = None

    def __copy__(self):
        raise Exception("Cannot copy state objects")

    def __deepcopy__(self):
        raise Exception("Cannot copy state objects")


class Action(metaclass=abc.ABCMeta):
    """Base class for all actions"""

    def __init__(self, name, analysis_state_manager=False):
        self.__name = name
        self.__analysis_state_manager = analysis_state_manager

    @property
    def name(self):
        return self.__name

    @property
    def analysis_state_manager(self):
        return self.__analysis_state_manager

    def _run(
        self,
        model,
        return_state=False,
        progress_bar=False,
        simulation_type=None,
        simulation_options=None,
    ):
        """Runs this Action on some input model and returns a solution.

        Parameters
        ----------
        model : Model
            Model to run this action on
        return_state : boolean
            If True the AnalysisState object is returned along with the solution
        progress_bar : bool, optional
            Whether to show progress bars or not
        simulation_options : dict
            Options for simulation to build and run

        Returns
        -------
        solution : BaseSolution
            Solution object generated by this action
        state : AnalysisState, when return_state = True
            The final state object after pasing through the action. This can be used
            to extract the models generated and tuned at later actions.
        """
        from .series import Series  # stop circular import

        before = finesse.config.show_progress_bars
        finesse.config.show_progress_bars = progress_bar

        state = AnalysisState(
            model,
            simulation_type=simulation_type,
            simulation_options=simulation_options,
        )
        try:
            if not self.analysis_state_manager:
                action = Series(self)
            else:
                action = self

            result = state.apply(action)

            if type(result) is tuple:
                sol = BaseSolution("root")
                for _ in result:
                    if _ is not None:
                        sol.add(_)
            else:
                sol = result

            if type(sol) is BaseSolution and len(sol.children) == 1:
                sol = sol[0]
        finally:
            state.finished()
            finesse.config.show_progress_bars = before

        if return_state:
            return sol, state
        else:
            return sol

    @abc.abstractmethod
    def _requests(self, model, memo, first=True):
        """Updates the memo dictionary with details about what this action needs from a
        simulation to run. Parent actions will get requests from all its child actions
        so that it can build a model that suits all of them, to minimise the amount of
        building.

        This method can do initial checks to make sure the model has the
        required features to perform the action too.

        memo['changing_parameters'] - append to this list the full name string
                                      of parameters that this action needs
        memo['input_nodes'] - append to this list the full-name string
                            of nodes that this action needs to keep. This should
                            list nodes that are inputted.
        memo['output_nodes'] - append to this list the full-name string
                            of nodes that this action needs to keep.
                            This should be used where actions are
                            accessing node outputs without using a
                            detector element (which registers that
                            nodes should be kept already).

        Parameters
        ----------
        model : Model
            The Model that the action will be operating on
        memo : defaultdict(list)
            A dictionary that should be filled with requests
        first : boolean
            True if this is the first request being made
        """
        raise NotImplementedError()

    def get_requests(self, model):
        memo = defaultdict(list)
        self._requests(model, memo)
        return memo

    @abc.abstractmethod
    def _do(self, state: AnalysisState) -> BaseSolution:
        pass

    def plan(self, previous=None):
        """Returns an expected plan for the actions that will be run in a tree form.
        This may not be exactly what is ran.

        Returns
        -------
        plan : TreeNode
        """
        if previous is None:
            previous = TreeNode("start")

        me = TreeNode(f"{self.name} - {self.__class__.__name__}")
        me.empty = not self.analysis_state_manager
        previous.add(me)

        found_actions = []

        for value in self.__dict__.values():
            if isinstance(value, Action):
                found_actions.append(value)
            elif isinstance(value, (tuple, list, set)):
                for _ in value:
                    if isinstance(_, Action):
                        found_actions.append(_)

        for action in found_actions:
            action.plan(me)
        return previous



import cython
import contextlib
import logging
import numpy as np

from libc.stdlib cimport free, calloc

from finesse.components.node import NodeType
from finesse.frequency cimport frequency_info_t, FrequencyContainer
from finesse.cymath.complex cimport complex_t

LOGGER = logging.getLogger(__name__)


cdef class HOMSolver(BaseSolver):
    """This is class provides an interface for generic simulations that are solving
    for a vector of higher order modes at each node. This allows detectors and other
    calculation code to be able to perform the same calculations without being
    specified. This class should be inherited to provide specific implementations.
    Considerations are:

        - HOM vector at each node should be contiguous in memory
        - Not all nodes will have a HOM vector if it isn't being solved for
        - Signal nodes will have a single "HOM"


    """
    def __init__(
        self,
        str name,
        list nodes,
        FrequencyContainer optical_frequencies,
        dict signal_frequencies,
        bint is_signal_matrix,
        bint forced_refill,
        dict node_aliases,
        int num_optical_homs,
        bint debug_mode = False,
    ):
        assert(num_optical_homs >= 1)
        self.nhoms = num_optical_homs
        super().__init__(
            name,
            nodes,
            optical_frequencies,
            signal_frequencies,
            is_signal_matrix,
            forced_refill,
            node_aliases,
            debug_mode=debug_mode
            )

        if is_signal_matrix:
            self._noise_matrices = dict()

    cpdef setup_nodes(self, list all_nodes, dict node_aliases):
        cdef:
            Py_ssize_t i, s_rhs_idx, s_f_idx, Nsm, Neq
            CNodeInfo *info
            frequency_info_t *finfo_ptr = NULL

        BaseSolver.setup_nodes(self, all_nodes, node_aliases)

        self._c_node_info = <CNodeInfo*> calloc(self.num_nodes, sizeof(CNodeInfo))
        if not self._c_node_info:
            raise MemoryError()

        s_rhs_idx = 0 # total number of states in the matrix so far
        s_f_idx = 0 # Number of frequency submatrices so far
        i = 0

        for n in self.nodes.values():
            if n in node_aliases:
                continue

            self.get_node_matrix_params(n, &Neq, &Nsm, &finfo_ptr)
            info = &self._c_node_info[i]

            info.index = self.node_2_index[n.full_name]
            info.rhs_index = s_rhs_idx
            info.freq_index = s_f_idx
            info.nfreqs = Nsm
            info.nhoms = Neq
            info.frequencies = finfo_ptr

            s_rhs_idx += Neq * Nsm  # Track how many equations we are going through
            s_f_idx   += Nsm        # keep track of how many frequencies*nodes
            i += 1

        self.out_view_size = s_rhs_idx

    cpdef construct(self):
        self.out_view = np.zeros(self.out_view_size, dtype=np.complex128)

    @contextlib.contextmanager
    def component_edge_fill(self, comp, edgestr, f1, f2, conjugate = False):
        """
        Returns a matrix for the submatrix an element has requested
        for different connections it needs. The key is::

            (element, connection_name, ifreq, ofreq)

        This is a context manager, to be used like with sim.component_edge_fill(key) as mat::

            mat[:] = computations

        Parameters
        ----------
        element : finesse.component.Connector
            The object reference that created the requests.
        connection_name : str
            String name given to the connection.
        ifreq : finesse.Frequency
            Incoming frequency.
        ofreq : finesse.Frequency
            Outgoing frequency.

        Returns
        -------
        matrix
        """
        #the two-stage nature of this will make some fill checks and hooks
        #much more safe (and powerfull)

        #this will be helpful for on-demand filling and will also help improve
        #make DC simulation of squeezers work (because post-fill transformations
        #will be needed)
        key = (comp, edgestr, f1.index, f2.index)
        mat = self._submatrices[key]
        yield mat
        #check things, or transform things here
        if conjugate:
            mat[:].imag *= -1
        return

    @contextlib.contextmanager
    def component_edge_fill3(self, owner_idx, conn_idx, f1_index, f2_index):
        if conn_idx < 0:
            raise IndexError(f"This connection was not included in the simulation. {owner_idx, conn_idx, f1_index, f2_index}")
        mat = self._submatrices[(owner_idx, conn_idx, f1_index, f2_index)]
        yield mat
        return

    cdef get_node_matrix_params(self, node, Py_ssize_t *Ns, Py_ssize_t *Nf, frequency_info_t** fptr):
        """For a given node in the simulation this should set the provided pointers
        regarding the number of states and submatricies that are required in the matrix:
            - Ns : Number of unique states at the node per frequency
            - Nf : Number of frequencies at the node
            - fptr : Pointer to frequency_info_t for details on the number of frequencies
        """
        assert(Ns)
        assert(Nf)
        assert(fptr)
        assert(self.nhoms > 0)
        assert(self.optical_frequencies.size > 0)

        cdef FrequencyContainer ficnt
        if node.type is NodeType.OPTICAL:
            Ns[0] = self.nhoms
            Nf[0] = self.optical_frequencies.size
            fptr[0] = &self.optical_frequencies.frequency_info[0]
        elif node.type is NodeType.MECHANICAL:
            # Higher order mechanical modes at a particular frequency. This should probably
            # be kept as 1 mode per frequency, additional mechanical degrees of freedom should
            # be defined as a separate node in a port.
            ficnt = <FrequencyContainer>(self.signal_frequencies[node])
            Ns[0] = 1
            Nf[0] = ficnt.size
            fptr[0] = &(ficnt.frequency_info[0])
        elif node.type is NodeType.ELECTRICAL:
            ficnt = <FrequencyContainer>(self.signal_frequencies[node])
            Ns[0] = 1 # no higher order modes of electronics as far as I'm aware...
            Nf[0] = ficnt.size
            fptr[0] = &(ficnt.frequency_info[0])
        else:
            raise Exception("Node type not handled")

    cpdef add_noise_matrix(self, object key):
        raise NotImplementedError

    cpdef set_source(self, object node, int freq_idx, int hom_idx, complex value):
        raise NotImplementedError()

    cdef int set_source_fast(self, Py_ssize_t node_id, Py_ssize_t freq_idx, Py_ssize_t hom_idx, complex_t value, Py_ssize_t rhs_index) except -1:
        return -1

    cdef int set_source_fast_2(self, Py_ssize_t rhs_idx, complex_t value) except -1:
        return -1

    cdef int set_source_fast_3(self, Py_ssize_t rhs_idx, complex_t value, Py_ssize_t rhs_index) except -1:
        return -1

    cpdef Py_ssize_t findex(self, object node, Py_ssize_t freq):
        return self.findex_fast(self.node_id(node), freq)

    cdef Py_ssize_t findex_fast(self, Py_ssize_t node_id, Py_ssize_t freq) nogil:
        assert self._c_node_info != NULL
        cdef:
            CNodeInfo ni = self._c_node_info[node_id]
            Py_ssize_t freq_idx = ni.freq_index

        return freq_idx + freq

    cpdef Py_ssize_t field(self, object node, Py_ssize_t freq=0, Py_ssize_t hom=0):

        """
        Returns simulation unique index of a field at a particular frequency
        index at this node.

        Parameters
        ----------
        node : :class:`.Node`
            Node object to get the index of.
        freq : int
            Frequency index.
        hom : int, optional
            Higher Order Mode index, defaults to zero.
        """
        return self.field_fast(self.node_id(node), freq, hom)

    cdef Py_ssize_t field_fast(self, Py_ssize_t node_id, Py_ssize_t freq=0, Py_ssize_t hom=0) noexcept nogil:
        if self._c_node_info == NULL:
            return -1
        cdef:
            CNodeInfo ni = self._c_node_info[node_id]
            Py_ssize_t Nh = ni.nhoms
            Py_ssize_t rhs_idx = ni.rhs_index
        return rhs_idx + freq * Nh + hom

    cdef inline Py_ssize_t field_fast_2(
        self,
        Py_ssize_t node_rhs_idx,
        Py_ssize_t num_hom,
        Py_ssize_t freq,
        Py_ssize_t hom) noexcept nogil:
        """Inlined function to return field index fast."""
        return node_rhs_idx + freq * num_hom + hom

    cpdef complex_t get_out(self, object node, Py_ssize_t freq=0, Py_ssize_t hom=0):
        return self.get_out_fast(self.node_id(node), freq, hom)

    cpdef get_node_info(self, node):
        """For a given node (object or name) the key parameters for where this node is
        represented in the matrix of linear equations.

        Parameters
        ----------
        node : [str | Node]
            The name or the Node object of the node.

        Returns
        -------
        dict: A dictionary containing the following information about the node:
            - index: The index of the node.
            - rhs_index: The index of the right-hand side vector associated with the node.
            - freq_index: The index of the frequency vector associated with the node.
            - nfreqs: The number of frequencies.
            - nhoms: The number of higher order modes. [TODO generalise to pixels/HOMs/whatever]
         """

        assert self._c_node_info != NULL
        cdef int i
        if type(node) is str:
            i = self.node_2_index[node]
        else:
            i = self.node_2_index[node.full_name]

        cdef CNodeInfo ni = self._c_node_info[i]
        return {
            "index": ni.index,
            "rhs_index": ni.rhs_index,
            "freq_index": ni.freq_index,
            "nfreqs": ni.nfreqs,
            "nhoms": ni.nhoms,
        }

    cpdef complex_t[::1] node_field_vector(self, node_id_str_object node, Py_ssize_t freq_idx):
        """
        Returns the higher order mode field vector of a given node at a specific
        frequency index.

        Parameters
        ----------
        node : [int|object|str]
            The node for which to retrieve the field vector. This can be a string full-name
            of a node, 'm1.p1.i', or a node object. It can also be an integer index of the
            node for this simulation.
        freq_idx : unsigned long
            The index of the frequency at which to retrieve the field vector.

        Returns
        -------
        np.ndarray:
            A copy of the field vector of the node at the specified frequency index.
        """
        cdef:
            Py_ssize_t N = 0
            Py_ssize_t i = 0
            complex_t *ptr = NULL

        if cython.int is node_id_str_object or cython.long is node_id_str_object:
            ptr = self.node_field_vector_fast(node, freq_idx, &N)
        else:
            i = self.node_id(node)
            ptr = self.node_field_vector_fast(i, freq_idx, &N)

        # TODO this is marked as slow in cython annotation, should check if we
        # can speed that up, maybe many cases do not need a copy but a view
        return (<complex_t[:N]>ptr)[:]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef complex_t* node_field_vector_fast(self, Py_ssize_t node_idx, Py_ssize_t freq_idx, Py_ssize_t *size) noexcept nogil:
        """
        Returns a pointer to the higher order mode field vector for a given node
        and frequency index.

        Parameters
        ----------
        node_idx : Py_ssize_t
            The index of the node.
        freq_idx : Py_ssize_t
            The index of the frequency.
        size : Py_ssize_t*
            A pointer to store the size of the field vector.

        Returns
        -------
        complex_t*
            A pointer to the field vector if the node index is valid and the frequency
            index is within range, otherwise returns NULL.
        """
        cdef Py_ssize_t idx

        if size != NULL:
            size[0] = self.nhoms

        if node_idx >= 0 and node_idx < self.num_nodes:
            idx = self._c_node_info[node_idx].rhs_index + freq_idx * self.nhoms
            if idx < self.out_view_size:
                return &self.out_view[idx]
            else:
                return NULL
        else:
            return NULL

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef complex_t get_out_fast(self, Py_ssize_t node_id, Py_ssize_t freq, Py_ssize_t hom) noexcept nogil:
        """
        Get the output value at a specific node, frequency, and higher order mode index.

        Parameters
        ----------
        node_id : Py_ssize_t
            The ID of the node.
        freq : Py_ssize_t, optional
            The frequency index.
        hom : Py_ssize_t, optional
            The higher order mode index.

        Returns
        -------
        complex_t
            The output value at the specified node, frequency, and homodyne index.
        """
        cdef Py_ssize_t field_idx = self.field_fast(node_id, freq, hom)
        return self.out_view[field_idx]

    def __dealloc__(self):
        if self._c_node_info != NULL:
            free(self._c_node_info)



from finesse.cymath.complex cimport complex_t
from finesse.cymath.gaussbeam cimport beam_param


cdef struct unm_workspace:
    double k
    double root_w0x # sqrt(w0x)
    double root_w0y # sqrt(w0y)
    double gouyx
    double gouyy
    double root2_on_wx # sqrt(2) / wx
    double root2_on_wy # sqrt(2) / wy
    complex_t inv_qx
    complex_t inv_qy
    complex_t xgauss_exponent # - ik / (2 qx)
    complex_t ygauss_exponent # - ik / (2 qy)
    complex_t z1x # sqrt(i Im{qx} / qx)
    complex_t z1y # sqrt(i Im{qy} / qy)
    complex_t z2x # sqrt(i Im(qx) qx* / (-i Im(qx) qx))
    complex_t z2y # sqrt(i Im(qy) qy* / (-i Im(qy) qy))
    # Used when n, m specified in unm_ws_recache so that more
    # optimised fixed mode index calculations can be performed
    int n, m
    complex_t xpre_factor
    complex_t ypre_factor

cdef void unm_ws_recache_from_bp(
    unm_workspace* uws,
    const beam_param* qx,
    const beam_param* qy,
    int* n=?,
    int* m=?,
) noexcept nogil

cdef void unm_ws_recache(
    unm_workspace* uws,
    complex_t qx,
    complex_t qy,
    double nr,
    double lambda0,
    int* n=?,
    int* m=?
) noexcept nogil

cdef struct unm_factor_store:
    # (2 / pi)^{1/4} * 1 / sqrt(2**n * n! * w0x) * z1x * z2x**m
    # for each mode index n
    complex_t* xpre_factor_cache
    # same as above but for y and each mode index m
    complex_t* ypre_factor_cache
    int nsize
    int msize

cdef void unm_factor_store_init(
    unm_factor_store* ufs,
    const unm_workspace* uws,
    int max_n,
    int max_m,
    bint remove_gouy,
    bint flip_odd_x_modes
) noexcept nogil

cdef void unm_factor_store_recache(
    unm_factor_store* ufs,
    const unm_workspace* uws,
    bint remove_gouy,
    bint flip_odd_x_modes
) noexcept nogil

cdef void unm_factor_store_free(unm_factor_store* ufs) noexcept nogil

cdef complex_t u_nm(const unm_workspace* uws, int n, int m, double x, double y) noexcept nogil
cdef complex_t u_nm__fast(
    const unm_workspace* uws,
    const unm_factor_store* ufs,
    int n, int m,
    double x, double y
) noexcept nogil
cdef complex_t u_nmconst(const unm_workspace* uws, double x, double y) noexcept nogil

cdef complex_t u_n(const unm_workspace* uws, int n, double x) noexcept nogil
cdef complex_t u_n__fast(
    const unm_workspace* uws,
    const unm_factor_store* ufs,
    int n,
    double x,
) noexcept nogil
cdef complex_t u_nconst(const unm_workspace* uws, double x) noexcept nogil

cdef complex_t u_m(const unm_workspace* uws, int m, double y) noexcept nogil
cdef complex_t u_m__fast(
    const unm_workspace* uws,
    const unm_factor_store* ufs,
    int m,
    double y,
) noexcept nogil
cdef complex_t u_mconst(const unm_workspace* uws, double y) noexcept nogil


cpdef Py_ssize_t field_index(int n, int m, const int[:, ::1] homs) noexcept nogil
cpdef bint in_mask(int n, int m, const int[:, ::1] mask) noexcept nogil


cdef class HGModeWorkspace:
    cdef public int n, m

    cdef readonly:
        complex_t qx, qy
        double nr
        double lambda0

        bint is_astigmatic

    cdef unm_workspace uws

    cpdef set_values(self, qx=?, qy=?, nr=?, lambda0=?)

    cpdef u_n(self, x, complex_t[::1] out=?)
    cpdef u_m(self, y, complex_t[::1] out=?)
    cpdef u_nm(self, x, y, out=?)
